#include <glib.h>

#ifndef BD_UTILS_SIZES
#define BD_UTILS_SIZES

#define KIBIBYTE *1024ULL
#define MEBIBYTE *1024ULL KIBIBYTE
#define GIBIBYTE *1024ULL MEBIBYTE
#define TEBIBYTE *1024ULL GIBIBYTE
#define PEBIBYTE *1024ULL TEBIBYTE
#define EXBIBYTE *1024ULL PEBIBYTE

#define KiB KIBIBYTE
#define MiB MEBIBYTE
#define GiB GIBIBYTE
#define TiB TEBIBYTE
#define PiB PEBIBYTE
#define EiB EXBIBYTE

#define KILOBYTE *1000ULL
#define MEGABYTE *1000ULL KILOBYTE
#define GIGABYTE *1000ULL MEGABYTE
#define TERABYTE *1000ULL GIGABYTE
#define PETABYTE *1000ULL TERABYTE
#define EXABYTE *1000ULL PETABYTE

#define KB KILOBYTE
#define MB MEGABYTE
#define GB GIGABYTE
#define TB TERABYTE
#define PB PETABYTE
#define EB EXABYTE

#endif /* BD_UTILS_SIZES */
